"""Web Push subscription management endpoints."""

from pathlib import Path

from litestar import Controller, Request, get, post
from litestar.response import File, Response
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.auth.session_keys import SESSION_USER_ID
from skrift.lib.push import get_vapid_public_key, remove_subscription, save_subscription

_SW_PATH = Path(__file__).parent.parent / "static" / "sw.js"


class PushController(Controller):
    """Endpoints for Web Push subscription management."""

    path = "/push"

    @get("/vapid-key")
    async def vapid_key(self, request: Request, db_session: AsyncSession) -> Response:
        """Return the VAPID public key for the frontend to subscribe."""
        public_key = await get_vapid_public_key(db_session)
        return Response(content={"publicKey": public_key})

    @post("/subscribe")
    async def subscribe(self, request: Request, db_session: AsyncSession) -> Response:
        """Register a push subscription for the current user."""
        user_id = request.session.get(SESSION_USER_ID)
        if not user_id:
            return Response(content={"error": "unauthorized"}, status_code=401)

        body = await request.json()
        endpoint = body.get("endpoint", "")
        keys = body.get("keys", {})
        p256dh = keys.get("p256dh", "")
        auth = keys.get("auth", "")

        if not endpoint or not p256dh or not auth:
            return Response(
                content={"error": "endpoint and keys (p256dh, auth) required"},
                status_code=400,
            )

        await save_subscription(db_session, user_id, endpoint, p256dh, auth)
        return Response(content={"ok": True}, status_code=201)

    @post("/unsubscribe")
    async def unsubscribe(self, request: Request, db_session: AsyncSession) -> Response:
        """Remove a push subscription."""
        user_id = request.session.get(SESSION_USER_ID)
        if not user_id:
            return Response(content={"error": "unauthorized"}, status_code=401)

        body = await request.json()
        endpoint = body.get("endpoint", "")
        if not endpoint:
            return Response(content={"error": "endpoint required"}, status_code=400)

        removed = await remove_subscription(db_session, endpoint)
        return Response(content={"ok": removed})


@get("/sw.js", media_type="application/javascript")
async def service_worker() -> File:
    """Serve the base service worker at root scope for push notifications."""
    return File(
        path=_SW_PATH,
        media_type="application/javascript",
        headers={"Service-Worker-Allowed": "/", "Cache-Control": "no-cache"},
    )



"""Local filesystem storage backend."""

from __future__ import annotations

import asyncio
import hashlib
from collections.abc import AsyncIterator
from pathlib import Path

from skrift.lib.storage.base import StoredFile


class LocalStorageBackend:
    """Store files on the local filesystem with content-hash subdirectories."""

    def __init__(self, base_path: Path, store_name: str = "default") -> None:
        self._base_path = base_path
        self._store_name = store_name

    async def put(self, key: str, data: bytes, content_type: str) -> StoredFile:
        path = self._key_to_path(key)
        await asyncio.to_thread(self._write_file, path, data)
        return StoredFile(
            key=key,
            url=self._build_url(key),
            content_type=content_type,
            size=len(data),
            content_hash=hashlib.sha256(data).hexdigest(),
        )

    async def get(self, key: str) -> bytes:
        path = self._key_to_path(key)
        return await asyncio.to_thread(path.read_bytes)

    async def delete(self, key: str) -> None:
        path = self._key_to_path(key)
        await asyncio.to_thread(self._unlink, path)

    async def exists(self, key: str) -> bool:
        path = self._key_to_path(key)
        return await asyncio.to_thread(path.exists)

    async def list_keys(self, prefix: str = "") -> AsyncIterator[str]:
        base = self._base_path
        for path in await asyncio.to_thread(self._walk, base):
            key = str(path.relative_to(base))
            if key.startswith(prefix):
                yield key

    async def get_url(self, key: str) -> str:
        return self._build_url(key)

    # -- internal helpers --

    def _key_to_path(self, key: str) -> Path:
        """Map a key to a content-hash based subdirectory path."""
        # Keys are expected to be content hashes (or hash-based); use first 4
        # chars for two levels of directory fan-out.
        if len(key) >= 4:
            return self._base_path / key[:2] / key[2:4] / key
        return self._base_path / key

    def _build_url(self, key: str) -> str:
        return f"/storage/{self._store_name}/{key}"

    @staticmethod
    def _write_file(path: Path, data: bytes) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(data)

    @staticmethod
    def _unlink(path: Path) -> None:
        path.unlink(missing_ok=True)

    @staticmethod
    def _walk(base: Path) -> list[Path]:
        if not base.exists():
            return []
        return [p for p in base.rglob("*") if p.is_file()]



"""On-demand image variant generation using Pillow."""

from __future__ import annotations

import io

from PIL import Image

IMAGE_SIZES: dict[str, tuple[int, int | None]] = {
    "icon": (64, 64),
    "thumb": (200, 200),
    "small": (400, None),
    "medium": (800, None),
    "cover": (1200, None),
    "og": (1200, 630),
}

_FORMAT_TO_CONTENT_TYPE = {
    "JPEG": "image/jpeg",
    "PNG": "image/png",
    "WEBP": "image/webp",
    "GIF": "image/gif",
}


def detect_image_content_type(data: bytes) -> str | None:
    """Detect image content type from magic bytes.

    Returns ``None`` if the data does not match a known image signature.
    """
    if data[:3] == b"\xff\xd8\xff":
        return "image/jpeg"
    if data[:8] == b"\x89PNG\r\n\x1a\n":
        return "image/png"
    if data[:4] == b"RIFF" and data[8:12] == b"WEBP":
        return "image/webp"
    if data[:6] in (b"GIF87a", b"GIF89a"):
        return "image/gif"
    return None


def resize_image(
    data: bytes,
    max_width: int,
    max_height: int | None,
) -> tuple[bytes, str]:
    """Resize an image to fit within the given dimensions.

    Preserves aspect ratio and original format. Does not upscale if the
    original is already smaller than the target.

    Returns:
        A tuple of ``(resized_bytes, content_type)``.
    """
    img = Image.open(io.BytesIO(data))
    orig_format = img.format or "PNG"
    content_type = _FORMAT_TO_CONTENT_TYPE.get(orig_format, "image/png")

    orig_w, orig_h = img.size

    if max_height is not None:
        # Fit within box — don't upscale
        if orig_w <= max_width and orig_h <= max_height:
            return data, content_type

        img.thumbnail((max_width, max_height), Image.LANCZOS)
    else:
        # Width-constrained only — don't upscale
        if orig_w <= max_width:
            return data, content_type

        ratio = max_width / orig_w
        new_h = int(orig_h * ratio)
        img = img.resize((max_width, new_h), Image.LANCZOS)

    buf = io.BytesIO()
    save_kwargs: dict = {}
    if orig_format == "JPEG":
        save_kwargs["quality"] = 85
        save_kwargs["optimize"] = True
    elif orig_format == "PNG":
        save_kwargs["optimize"] = True
    elif orig_format == "WEBP":
        save_kwargs["quality"] = 85

    img.save(buf, format=orig_format, **save_kwargs)
    return buf.getvalue(), content_type


def variant_filename(key: str, size_name: str) -> str:
    """Derive the on-disk filename for a sized variant.

    E.g. ``variant_filename("abc123def.jpg", "thumb")`` → ``"abc123def.jpg.thumb"``
    """
    return f"{key}.{size_name}"



"""ASGI middleware for serving locally-stored assets."""

from __future__ import annotations

import logging
import mimetypes
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import parse_qs

from litestar.types import ASGIApp, Receive, Scope, Send

from skrift.lib.imaging import IMAGE_SIZES, detect_image_content_type, resize_image, variant_filename
from skrift.middleware.helpers import send_not_found

if TYPE_CHECKING:
    from skrift.config import StorageConfig

logger = logging.getLogger(__name__)


class StorageFilesMiddleware:
    """Serve files from local storage backends at ``/storage/{store}/{key}``.

    Only handles stores configured with ``backend = "local"``.  Remote backends
    (S3, etc.) serve via their own URLs and are not intercepted here.

    Supports ``?size=name`` query parameter for on-demand image resizing.
    Resized variants are cached alongside originals for subsequent requests.
    """

    def __init__(self, app: ASGIApp, storage_config: StorageConfig) -> None:
        self.app = app
        self._storage_config = storage_config

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http" or not scope["path"].startswith("/storage/"):
            await self.app(scope, receive, send)
            return

        # Parse /storage/{store_name}/{key...}
        rest = scope["path"][len("/storage/"):]
        slash_idx = rest.find("/")
        if slash_idx == -1 or not rest[:slash_idx]:
            await send_not_found(send)
            return

        store_name = rest[:slash_idx]
        key = rest[slash_idx + 1:]

        if not key:
            await send_not_found(send)
            return

        # Only serve local backends
        store_cfg = self._storage_config.stores.get(store_name)
        if store_cfg is None or store_cfg.backend != "local":
            await send_not_found(send)
            return

        # Security: reject traversal and null bytes
        if ".." in key.split("/") or "\x00" in key:
            await send_not_found(send)
            return

        base_path = Path(store_cfg.local_path).resolve()

        # Reconstruct the on-disk path through LocalStorageBackend's layout
        if len(key) >= 4:
            candidate = base_path / key[:2] / key[2:4] / key
        else:
            candidate = base_path / key

        try:
            resolved = candidate.resolve()
        except (OSError, ValueError):
            await send_not_found(send)
            return

        if not resolved.is_relative_to(base_path):
            await send_not_found(send)
            return

        if not resolved.is_file():
            await send_not_found(send)
            return

        # Check for ?size=name variant request
        qs = scope.get("query_string", b"")
        params = parse_qs(qs.decode("latin-1") if isinstance(qs, bytes) else qs)
        size_name = params.get("size", [None])[0]

        result = None
        if size_name and size_name in IMAGE_SIZES:
            result = self._get_or_create_variant(resolved, size_name)

        if result is not None:
            content, media_type = result
        else:
            content = resolved.read_bytes()
            media_type = mimetypes.guess_type(str(resolved))[0] or "application/octet-stream"

        await send({
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", media_type.encode()),
                (b"content-length", str(len(content)).encode()),
            ],
        })
        await send({"type": "http.response.body", "body": content})

    @staticmethod
    def _get_or_create_variant(original: Path, size_name: str) -> tuple[bytes, str] | None:
        """Return cached variant bytes or generate and cache them.

        The variant file is stored alongside the original with a ``.{size_name}``
        suffix (e.g. ``abc123def456.thumb``).

        Returns ``None`` if the original is not a recognized image format.
        """
        variant_name = variant_filename(original.name, size_name)
        variant_path = original.parent / variant_name

        if variant_path.is_file():
            content = variant_path.read_bytes()
            ct = detect_image_content_type(content) or "application/octet-stream"
            return content, ct

        original_bytes = original.read_bytes()

        # Skip non-image files
        if detect_image_content_type(original_bytes) is None:
            return None

        max_w, max_h = IMAGE_SIZES[size_name]
        resized_bytes, content_type = resize_image(original_bytes, max_w, max_h)

        try:
            variant_path.write_bytes(resized_bytes)
        except OSError:
            logger.warning("Failed to cache variant %s", variant_path, exc_info=True)

        return resized_bytes, content_type



"""Pluggable multi-backend asset storage."""

from skrift.lib.storage.base import StorageBackend, StoredFile
from skrift.lib.storage.local import LocalStorageBackend
from skrift.lib.storage.manager import StorageManager

__all__ = ["LocalStorageBackend", "StorageBackend", "StorageManager", "StoredFile"]



"""OAuth2 client management admin controller."""

from __future__ import annotations

from uuid import UUID

from litestar import Controller, Request, get, post
from litestar.response import Template as TemplateResponse, Redirect
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.auth.guards import auth_guard, Permission
from skrift.auth.scopes import SCOPE_DEFINITIONS
from skrift.admin.helpers import get_admin_context
from skrift.admin.navigation import ADMIN_NAV_TAG
from skrift.db.models.oauth2_client import OAuth2Client
from skrift.db.services import oauth2_service
from skrift.lib.flash import flash_error, flash_success, get_flash_messages


class OAuth2ClientAdminController(Controller):
    """Controller for managing OAuth2 clients in admin."""

    path = "/admin"
    guards = [auth_guard]

    @get(
        "/oauth-clients",
        tags=[ADMIN_NAV_TAG],
        guards=[auth_guard, Permission("manage-oauth-clients")],
        opt={"label": "OAuth Clients", "icon": "key", "order": 90},
    )
    async def list_clients(
        self, request: Request, db_session: AsyncSession
    ) -> TemplateResponse:
        """List all OAuth2 clients."""
        ctx = await get_admin_context(request, db_session)
        clients = await oauth2_service.list_clients(db_session)
        flash_messages = get_flash_messages(request)
        return TemplateResponse(
            "admin/oauth2/list.html",
            context={"flash_messages": flash_messages, "clients": clients, **ctx},
        )

    @get(
        "/oauth-clients/new",
        guards=[auth_guard, Permission("manage-oauth-clients")],
    )
    async def new_client_form(
        self, request: Request, db_session: AsyncSession
    ) -> TemplateResponse:
        """Show create client form."""
        ctx = await get_admin_context(request, db_session)
        flash_messages = get_flash_messages(request)
        return TemplateResponse(
            "admin/oauth2/edit.html",
            context={
                "flash_messages": flash_messages,
                "client": None,
                "available_scopes": SCOPE_DEFINITIONS,
                **ctx,
            },
        )

    @post(
        "/oauth-clients/new",
        guards=[auth_guard, Permission("manage-oauth-clients")],
    )
    async def create_client(
        self, request: Request, db_session: AsyncSession
    ) -> Redirect:
        """Create a new OAuth2 client."""
        form_data = await request.form()
        display_name = form_data.get("display_name", "").strip()

        if not display_name:
            flash_error(request, "Display name is required")
            return Redirect(path="/admin/oauth-clients/new")

        redirect_uris_raw = form_data.get("redirect_uris", "")
        redirect_uris = [u.strip() for u in redirect_uris_raw.split("\n") if u.strip()]

        allowed_scopes = []
        for scope_name in SCOPE_DEFINITIONS:
            if form_data.get(f"scope_{scope_name}") == "on":
                allowed_scopes.append(scope_name)

        client = await oauth2_service.create_client(
            db_session, display_name, redirect_uris, allowed_scopes
        )

        request.session["new_secret"] = client.client_secret
        flash_success(request, f"Client '{display_name}' created")
        return Redirect(path=f"/admin/oauth-clients/{client.id}/edit")

    @get(
        "/oauth-clients/{client_db_id:uuid}/edit",
        guards=[auth_guard, Permission("manage-oauth-clients")],
    )
    async def edit_client_form(
        self, request: Request, db_session: AsyncSession, client_db_id: UUID
    ) -> TemplateResponse:
        """Show edit client form."""
        ctx = await get_admin_context(request, db_session)
        result = await db_session.execute(
            select(OAuth2Client).where(OAuth2Client.id == client_db_id)
        )
        client = result.scalar_one_or_none()
        if not client:
            flash_error(request, "Client not found")
            return Redirect(path="/admin/oauth-clients")

        new_secret = request.session.pop("new_secret", None)
        flash_messages = get_flash_messages(request)
        return TemplateResponse(
            "admin/oauth2/edit.html",
            context={
                "flash_messages": flash_messages,
                "client": client,
                "available_scopes": SCOPE_DEFINITIONS,
                "new_secret": new_secret,
                **ctx,
            },
        )

    @post(
        "/oauth-clients/{client_db_id:uuid}/edit",
        guards=[auth_guard, Permission("manage-oauth-clients")],
    )
    async def update_client(
        self, request: Request, db_session: AsyncSession, client_db_id: UUID
    ) -> Redirect:
        """Update an existing OAuth2 client."""
        result = await db_session.execute(
            select(OAuth2Client).where(OAuth2Client.id == client_db_id)
        )
        client = result.scalar_one_or_none()
        if not client:
            flash_error(request, "Client not found")
            return Redirect(path="/admin/oauth-clients")

        form_data = await request.form()
        display_name = form_data.get("display_name", "").strip()
        redirect_uris_raw = form_data.get("redirect_uris", "")
        redirect_uris = [u.strip() for u in redirect_uris_raw.split("\n") if u.strip()]
        is_active = form_data.get("is_active") == "on"

        allowed_scopes = []
        for scope_name in SCOPE_DEFINITIONS:
            if form_data.get(f"scope_{scope_name}") == "on":
                allowed_scopes.append(scope_name)

        await oauth2_service.update_client(
            db_session, client,
            display_name=display_name,
            redirect_uris=redirect_uris,
            allowed_scopes=allowed_scopes,
            is_active=is_active,
        )

        flash_success(request, f"Client '{display_name}' updated")
        return Redirect(path="/admin/oauth-clients")

    @post(
        "/oauth-clients/{client_db_id:uuid}/delete",
        guards=[auth_guard, Permission("manage-oauth-clients")],
    )
    async def delete_client(
        self, request: Request, db_session: AsyncSession, client_db_id: UUID
    ) -> Redirect:
        """Delete an OAuth2 client."""
        await oauth2_service.delete_client(db_session, client_db_id)
        flash_success(request, "Client deleted")
        return Redirect(path="/admin/oauth-clients")

    @post(
        "/oauth-clients/{client_db_id:uuid}/regenerate-secret",
        guards=[auth_guard, Permission("manage-oauth-clients")],
    )
    async def regenerate_secret(
        self, request: Request, db_session: AsyncSession, client_db_id: UUID
    ) -> Redirect:
        """Regenerate a client's secret."""
        result = await db_session.execute(
            select(OAuth2Client).where(OAuth2Client.id == client_db_id)
        )
        client = result.scalar_one_or_none()
        if not client:
            flash_error(request, "Client not found")
            return Redirect(path="/admin/oauth-clients")

        new_secret = await oauth2_service.regenerate_client_secret(db_session, client)
        request.session["new_secret"] = new_secret
        flash_success(request, f"Secret regenerated for '{client.display_name}'")
        return Redirect(path=f"/admin/oauth-clients/{client_db_id}/edit")



"""Add OAuth2 server tables (oauth2_clients, revoked_tokens).

Revision ID: s1t2u3v4w5x6
Revises: m7n8o9p0q1r2
Create Date: 2026-03-04
"""

from alembic import op
import sqlalchemy as sa

revision = "s1t2u3v4w5x6"
down_revision = "m7n8o9p0q1r2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "oauth2_clients",
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("client_id", sa.String(255), nullable=False),
        sa.Column("client_secret", sa.String(255), nullable=False, server_default=""),
        sa.Column("display_name", sa.String(255), nullable=False),
        sa.Column("redirect_uris", sa.Text(), nullable=False, server_default=""),
        sa.Column("allowed_scopes", sa.Text(), nullable=False, server_default=""),
        sa.Column("is_active", sa.Boolean(), nullable=False, server_default=sa.text("true")),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()),
        sa.Column("sa_orm_sentinel", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_oauth2_clients_client_id", "oauth2_clients", ["client_id"], unique=True)

    op.create_table(
        "revoked_tokens",
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("jti", sa.String(64), nullable=False),
        sa.Column("token_type", sa.String(20), nullable=False),
        sa.Column("expires_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()),
        sa.Column("sa_orm_sentinel", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_revoked_tokens_jti", "revoked_tokens", ["jti"], unique=True)
    op.create_index("ix_revoked_tokens_expires_at", "revoked_tokens", ["expires_at"])


def downgrade() -> None:
    op.drop_index("ix_revoked_tokens_expires_at", table_name="revoked_tokens")
    op.drop_index("ix_revoked_tokens_jti", table_name="revoked_tokens")
    op.drop_table("revoked_tokens")
    op.drop_index("ix_oauth2_clients_client_id", table_name="oauth2_clients")
    op.drop_table("oauth2_clients")



"""Shared helpers for ASGI middleware."""

from litestar.types import Send


async def send_not_found(send: Send) -> None:
    """Send a plain-text 404 response."""
    await send({
        "type": "http.response.start",
        "status": 404,
        "headers": [(b"content-type", b"text/plain")],
    })
    await send({"type": "http.response.body", "body": b"Not Found"})



from datetime import datetime

from sqlalchemy import DateTime, Index, String, Text, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column

from skrift.db.base import Base


class PushSubscription(Base):
    """Web Push subscription for a user's browser."""

    __tablename__ = "push_subscriptions"

    user_id: Mapped[str] = mapped_column(String(255), nullable=False)
    endpoint: Mapped[str] = mapped_column(Text, nullable=False)
    key_p256dh: Mapped[str] = mapped_column(String(255), nullable=False)
    key_auth: Mapped[str] = mapped_column(String(255), nullable=False)
    last_used_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )

    __table_args__ = (
        UniqueConstraint("endpoint", name="uq_push_subscriptions_endpoint"),
        Index("ix_push_subscriptions_user_id", "user_id"),
    )



"""Shared helpers for public-facing controllers."""

from uuid import UUID

from litestar import Request
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.auth.session_keys import SESSION_USER_ID
from skrift.db.models.user import User
from skrift.db.services.setting_service import get_cached_site_theme
from skrift.lib.hooks import RESOLVE_THEME, apply_filters


async def get_user_context(request: Request, db_session: AsyncSession) -> dict:
    """Get user data for template context if logged in."""
    user_id = request.session.get(SESSION_USER_ID)
    if not user_id:
        return {"user": None}

    result = await db_session.execute(select(User).where(User.id == UUID(user_id)))
    user = result.scalar_one_or_none()
    return {"user": user}


async def resolve_theme(request: Request) -> str:
    """Resolve the active theme for this request via filter hook."""
    theme_name = get_cached_site_theme()
    return await apply_filters(RESOLVE_THEME, theme_name, request)



"""Shared page orchestration for admin controllers."""

from __future__ import annotations

from collections.abc import Collection
from datetime import UTC, datetime
from uuid import UUID

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from skrift.admin.helpers import PageFormData
from skrift.db.models import Page
from skrift.db.services import page_service
from skrift.db.services.asset_service import sync_page_assets


def _resolve_featured_asset_id(featured_asset_id: str | None) -> UUID | None:
    return UUID(featured_asset_id) if featured_asset_id else None


def _resolve_asset_ids(asset_ids: list[str]) -> list[UUID]:
    return [UUID(asset_id) for asset_id in asset_ids]


def _get_permissions(permission_source: Collection[str] | object) -> Collection[str]:
    permissions = getattr(permission_source, "permissions", permission_source)
    if not isinstance(permissions, Collection):
        raise TypeError("permissions must be a collection of strings")
    return permissions


async def list_pages_for_admin(
    db_session: AsyncSession,
    *,
    page_type_name: str,
    user_id: UUID,
    permissions: Collection[str] | object,
    manage_permission: str,
) -> list[Page]:
    """List admin pages for the current user, applying ownership rules."""
    query = (
        select(Page)
        .where(Page.type == page_type_name)
        .options(selectinload(Page.user))
        .order_by(Page.order.asc(), Page.created_at.desc())
    )

    permission_values = _get_permissions(permissions)
    if "administrator" not in permission_values and manage_permission not in permission_values:
        query = query.where(Page.user_id == user_id)

    result = await db_session.execute(query)
    return list(result.scalars().all())


async def create_typed_page(
    db_session: AsyncSession,
    *,
    form: PageFormData,
    user_id: UUID,
    page_type_name: str,
) -> Page:
    """Create a page and synchronize attached assets."""
    page = await page_service.create_page(
        db_session,
        slug=form.slug,
        title=form.title,
        content=form.content,
        is_published=form.is_published,
        published_at=datetime.now(UTC) if form.is_published else None,
        order=form.order,
        publish_at=form.publish_at,
        meta_description=form.meta_description,
        og_title=form.og_title,
        og_description=form.og_description,
        og_image=form.og_image,
        meta_robots=form.meta_robots,
        user_id=user_id,
        page_type=page_type_name,
        featured_asset_id=_resolve_featured_asset_id(form.featured_asset_id),
    )

    if form.asset_ids:
        await sync_page_assets(db_session, page.id, _resolve_asset_ids(form.asset_ids))

    return page


async def update_typed_page(
    db_session: AsyncSession,
    *,
    page: Page,
    form: PageFormData,
    user_id: UUID | None,
    page_type_name: str,
) -> Page | None:
    """Update a page and synchronize attached assets."""
    published_at = page.published_at
    if form.is_published and not page.is_published:
        published_at = datetime.now(UTC)

    updated_page = await page_service.update_page(
        db_session,
        page_id=page.id,
        slug=form.slug,
        title=form.title,
        content=form.content,
        is_published=form.is_published,
        published_at=published_at,
        order=form.order,
        publish_at=form.publish_at,
        meta_description=form.meta_description,
        og_title=form.og_title,
        og_description=form.og_description,
        og_image=form.og_image,
        meta_robots=form.meta_robots,
        page_type=page_type_name,
        user_id=user_id,
        featured_asset_id=_resolve_featured_asset_id(form.featured_asset_id),
    )

    await sync_page_assets(db_session, page.id, _resolve_asset_ids(form.asset_ids))
    return updated_page



"""Add page_assets association table.

Revision ID: a1b2c3d4e5f6
Revises: g9h0i1j2k3l4
Create Date: 2026-02-27
"""

from alembic import op
import sqlalchemy as sa

revision = "a1b2c3d4e5f6"
down_revision = "g9h0i1j2k3l4"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "page_assets",
        sa.Column("page_id", sa.Uuid(), nullable=False),
        sa.Column("asset_id", sa.Uuid(), nullable=False),
        sa.PrimaryKeyConstraint("page_id", "asset_id"),
        sa.ForeignKeyConstraint(["page_id"], ["pages.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["asset_id"], ["assets.id"], ondelete="CASCADE"),
    )


def downgrade() -> None:
    op.drop_table("page_assets")



"""SEO utilities for generating page metadata."""

from dataclasses import dataclass
from html import escape
from typing import TYPE_CHECKING

from markupsafe import Markup

from skrift.lib.hooks import hooks, PAGE_SEO_META, PAGE_OG_META

if TYPE_CHECKING:
    from skrift.db.models import Page


def _ensure_https(url: str) -> str:
    """Upgrade http:// to https:// for public-facing URLs."""
    if url.startswith("http://"):
        return "https://" + url[7:]
    return url


def _meta_tag(name: str, content: str) -> str:
    return f'<meta name="{escape(name)}" content="{escape(content)}">'


def _og_tag(prop: str, content: str) -> str:
    return f'<meta property="{escape(prop)}" content="{escape(content)}">'


@dataclass
class SEOMeta:
    """Standard SEO metadata for a page."""

    title: str
    description: str | None
    canonical_url: str
    robots: str | None

    def __html__(self) -> str:
        parts: list[str] = []
        if self.description:
            parts.append(_meta_tag("description", self.description))
        if self.robots:
            parts.append(_meta_tag("robots", self.robots))
        if self.canonical_url:
            parts.append(f'<link rel="canonical" href="{escape(self.canonical_url)}">')
        return Markup("\n    ".join(parts))


@dataclass
class OpenGraphMeta:
    """OpenGraph metadata for social sharing."""

    title: str
    description: str | None
    image: str | None
    url: str
    site_name: str
    type: str = "website"
    twitter_card: str = "summary_large_image"

    def __html__(self) -> str:
        parts = [
            _og_tag("og:title", self.title),
            _og_tag("og:type", self.type),
            _og_tag("og:url", self.url),
            _og_tag("og:site_name", self.site_name),
        ]
        if self.description:
            parts.append(_og_tag("og:description", self.description))
        if self.image:
            parts.append(_og_tag("og:image", self.image))
        parts.append(_meta_tag("twitter:card", self.twitter_card))
        if self.image:
            parts.append(_meta_tag("twitter:image", self.image))
        return Markup("\n    ".join(parts))


async def get_page_seo_meta(
    page: "Page",
    site_name: str,
    base_url: str,
) -> SEOMeta:
    """Generate SEO metadata for a page.

    Args:
        page: The page to generate metadata for
        site_name: The site name for title suffix
        base_url: Base URL for canonical URL generation

    Returns:
        SEOMeta dataclass with the metadata
    """
    base_url = _ensure_https(base_url)

    # Build canonical URL
    slug = page.slug.strip("/")
    canonical_url = f"{base_url.rstrip('/')}/{slug}" if slug else base_url.rstrip("/")

    # Build title with site name suffix
    title = f"{page.title} | {site_name}" if site_name else page.title

    meta = SEOMeta(
        title=title,
        description=page.meta_description,
        canonical_url=canonical_url,
        robots=page.meta_robots,
    )

    # Apply filter for extensibility
    meta = await hooks.apply_filters(PAGE_SEO_META, meta, page, site_name, base_url)

    return meta


async def get_page_og_meta(
    page: "Page",
    site_name: str,
    base_url: str,
    featured_image_url: str | None = None,
) -> OpenGraphMeta:
    """Generate OpenGraph metadata for a page.

    Args:
        page: The page to generate metadata for
        site_name: The site name
        base_url: Base URL for URL generation
        featured_image_url: Optional fallback image URL from the page's featured asset

    Returns:
        OpenGraphMeta dataclass with the metadata
    """
    base_url = _ensure_https(base_url)

    # Build URL
    slug = page.slug.strip("/")
    url = f"{base_url.rstrip('/')}/{slug}" if slug else base_url.rstrip("/")

    # Use og_* fields if set, otherwise fall back to page fields
    og_title = page.og_title or page.title
    og_description = page.og_description or page.meta_description
    image = page.og_image or featured_image_url
    if image and not image.startswith(("http://", "https://")):
        image = f"{base_url.rstrip('/')}/{image.lstrip('/')}"
    elif image:
        image = _ensure_https(image)

    meta = OpenGraphMeta(
        title=og_title,
        description=og_description,
        image=image,
        url=url,
        site_name=site_name,
    )

    # Apply filter for extensibility
    meta = await hooks.apply_filters(PAGE_OG_META, meta, page, site_name, base_url)

    return meta



"""Dynamic controller factory for per-type page admin sections."""

from __future__ import annotations

import logging
from datetime import UTC, datetime
from typing import Annotated
from uuid import UUID

from litestar import Controller, Request, get, post
from litestar.enums import RequestEncodingType
from litestar.params import Body
from litestar.response import Redirect, Template as TemplateResponse
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.admin.helpers import (
    check_page_access,
    extract_page_form_data,
    get_admin_context,
)
from skrift.admin.page_operations import (
    create_typed_page,
    list_pages_for_admin,
    update_typed_page,
)
from skrift.auth.session_keys import SESSION_USER_ID
from skrift.admin.navigation import ADMIN_NAV_TAG
from skrift.auth.guards import OwnerOrPermission, Permission, auth_guard
from skrift.auth.roles import permissions_for_type
from skrift.config import PageTypeConfig
from skrift.db.services import page_service, revision_service
from skrift.db.services.asset_service import get_asset_url
from skrift.lib.flash import flash_error, flash_success, get_flash_messages
from skrift.lib.storage import StorageManager

logger = logging.getLogger(__name__)


def create_page_type_controller(page_type: PageTypeConfig) -> type[Controller]:
    """Create a Controller subclass for a specific page type.

    Each generated controller:
    - Routes under /admin/{plural}
    - Filters all queries by Page.type == type_name
    - Uses type-specific permissions
    - Has its own admin nav entry
    """
    type_name = page_type.name          # "post"
    plural = page_type.plural            # "posts"
    icon = page_type.icon                # "pen-tool"
    nav_order = page_type.nav_order      # 10
    label = type_name.title()            # "Post"
    label_plural = plural.title()        # "Posts"
    admin_base = f"/admin/{plural}"      # "/admin/posts"
    perms = permissions_for_type(plural)  # {"manage": "manage-posts", ...}

    page_type_ctx = {
        "page_type_name": type_name,
        "page_type_plural": plural,
        "page_type_label": label,
        "page_type_label_plural": label_plural,
        "admin_type_base": admin_base,
    }

    def _get_page_or_redirect(page, request):
        if page is None:
            flash_error(request, f"{label} not found")
            return Redirect(path=admin_base)
        return None

    class _PageTypeController(Controller):
        path = "/admin"
        guards = [auth_guard]

        @get(
            f"/{plural}",
            tags=[ADMIN_NAV_TAG],
            guards=[auth_guard, OwnerOrPermission(perms["edit_own"], perms["manage"])],
            opt={"label": label_plural, "icon": icon, "order": nav_order},
        )
        async def list_pages(
            self, request: Request, db_session: AsyncSession
        ) -> TemplateResponse:
            ctx = await get_admin_context(request, db_session)
            pages = await list_pages_for_admin(
                db_session,
                page_type_name=type_name,
                user_id=UUID(request.session[SESSION_USER_ID]),
                permissions=ctx["permissions"],
                manage_permission=perms["manage"],
            )

            flash_messages = get_flash_messages(request)
            return TemplateResponse(
                "admin/pages/list.html",
                context={
                    "flash_messages": flash_messages,
                    "pages": pages,
                    **page_type_ctx,
                    **ctx,
                },
            )

        @get(
            f"/{plural}/new",
            guards=[auth_guard, OwnerOrPermission(perms["create"], perms["manage"])],
        )
        async def new_page(
            self, request: Request, db_session: AsyncSession
        ) -> TemplateResponse:
            ctx = await get_admin_context(request, db_session)
            flash_messages = get_flash_messages(request)
            return TemplateResponse(
                "admin/pages/edit.html",
                context={
                    "flash_messages": flash_messages,
                    "page": None,
                    "page_assets": [],
                    "asset_urls": {},
                    **page_type_ctx,
                    **ctx,
                },
            )

        @post(
            f"/{plural}/new",
            guards=[auth_guard, OwnerOrPermission(perms["create"], perms["manage"])],
        )
        async def create_page(
            self,
            request: Request,
            db_session: AsyncSession,
            data: Annotated[dict, Body(media_type=RequestEncodingType.URL_ENCODED)],
        ) -> Redirect:
            try:
                form = extract_page_form_data(data)
            except ValueError as e:
                flash_error(request, str(e))
                return Redirect(path=f"{admin_base}/new")

            if not form.title or not form.slug:
                flash_error(request, "Title and slug are required")
                return Redirect(path=f"{admin_base}/new")

            user_id = UUID(request.session[SESSION_USER_ID])

            try:
                page = await create_typed_page(
                    db_session,
                    user_id=user_id,
                    form=form,
                    page_type_name=type_name,
                )
                flash_success(request, f"{label} '{form.title}' created successfully!")
                return Redirect(path=admin_base)
            except Exception:
                logger.exception("Admin %s create failed", type_name)
                flash_error(
                    request,
                    f"Could not create {label.lower()}. Check the server logs and try again.",
                )
                return Redirect(path=f"{admin_base}/new")

        @get(
            f"/{plural}/{{page_id:uuid}}/edit",
            guards=[auth_guard, OwnerOrPermission(perms["edit_own"], perms["manage"])],
        )
        async def edit_page(
            self, request: Request, db_session: AsyncSession, page_id: UUID
        ) -> TemplateResponse:
            ctx = await get_admin_context(request, db_session)

            page = await page_service.get_page_by_id(db_session, page_id)
            if redirect := _get_page_or_redirect(page, request):
                return redirect

            await check_page_access(
                db_session, request, page, perms["edit_own"], perms["manage"]
            )

            # Resolve asset URLs for attached assets
            storage: StorageManager = request.app.state.storage_manager
            asset_urls = {
                str(asset.id): await get_asset_url(storage, asset)
                for asset in page.assets
            }

            flash_messages = get_flash_messages(request)
            return TemplateResponse(
                "admin/pages/edit.html",
                context={
                    "flash_messages": flash_messages,
                    "page": page,
                    "page_assets": page.assets,
                    "asset_urls": asset_urls,
                    **page_type_ctx,
                    **ctx,
                },
            )

        @post(
            f"/{plural}/{{page_id:uuid}}/edit",
            guards=[auth_guard, OwnerOrPermission(perms["edit_own"], perms["manage"])],
        )
        async def update_page(
            self,
            request: Request,
            db_session: AsyncSession,
            page_id: UUID,
            data: Annotated[dict, Body(media_type=RequestEncodingType.URL_ENCODED)],
        ) -> Redirect:
            try:
                form = extract_page_form_data(data)
            except ValueError as e:
                flash_error(request, str(e))
                return Redirect(path=f"{admin_base}/{page_id}/edit")

            if not form.title or not form.slug:
                flash_error(request, "Title and slug are required")
                return Redirect(path=f"{admin_base}/{page_id}/edit")

            page = await page_service.get_page_by_id(db_session, page_id)
            if redirect := _get_page_or_redirect(page, request):
                return redirect

            await check_page_access(
                db_session, request, page, perms["edit_own"], perms["manage"]
            )

            try:
                await update_typed_page(
                    db_session,
                    page=page,
                    form=form,
                    user_id=UUID(request.session[SESSION_USER_ID]),
                    page_type_name=type_name,
                )
                flash_success(request, f"{label} '{form.title}' updated successfully!")
                return Redirect(path=admin_base)
            except Exception:
                logger.exception("Admin %s update failed", type_name)
                flash_error(
                    request,
                    f"Could not update {label.lower()}. Check the server logs and try again.",
                )
                return Redirect(path=f"{admin_base}/{page_id}/edit")

        @post(
            f"/{plural}/{{page_id:uuid}}/publish",
            guards=[auth_guard, Permission(perms["manage"])],
        )
        async def publish_page(
            self, request: Request, db_session: AsyncSession, page_id: UUID
        ) -> Redirect:
            page = await page_service.get_page_by_id(db_session, page_id)
            if redirect := _get_page_or_redirect(page, request):
                return redirect

            await page_service.update_page(
                db_session,
                page_id=page_id,
                is_published=True,
                published_at=datetime.now(UTC),
            )

            flash_success(request, f"'{page.title}' has been published")
            return Redirect(path=admin_base)

        @post(
            f"/{plural}/{{page_id:uuid}}/unpublish",
            guards=[auth_guard, Permission(perms["manage"])],
        )
        async def unpublish_page(
            self, request: Request, db_session: AsyncSession, page_id: UUID
        ) -> Redirect:
            page = await page_service.get_page_by_id(db_session, page_id)
            if redirect := _get_page_or_redirect(page, request):
                return redirect

            await page_service.update_page(
                db_session,
                page_id=page_id,
                is_published=False,
            )

            flash_success(request, f"'{page.title}' has been unpublished")
            return Redirect(path=admin_base)

        @post(
            f"/{plural}/{{page_id:uuid}}/delete",
            guards=[auth_guard, OwnerOrPermission(perms["delete_own"], perms["manage"])],
        )
        async def delete_page(
            self, request: Request, db_session: AsyncSession, page_id: UUID
        ) -> Redirect:
            page = await page_service.get_page_by_id(db_session, page_id)
            if redirect := _get_page_or_redirect(page, request):
                return redirect

            await check_page_access(
                db_session, request, page, perms["delete_own"], perms["manage"]
            )

            page_title = page.title
            await page_service.delete_page(db_session, page_id)

            flash_success(request, f"'{page_title}' has been deleted")
            return Redirect(path=admin_base)

        @get(
            f"/{plural}/{{page_id:uuid}}/revisions",
            guards=[auth_guard, OwnerOrPermission(perms["edit_own"], perms["manage"])],
        )
        async def list_revisions(
            self, request: Request, db_session: AsyncSession, page_id: UUID
        ) -> TemplateResponse:
            ctx = await get_admin_context(request, db_session)

            page = await page_service.get_page_by_id(db_session, page_id)
            if redirect := _get_page_or_redirect(page, request):
                return redirect

            await check_page_access(
                db_session, request, page, perms["edit_own"], perms["manage"]
            )

            revisions = await revision_service.list_revisions(db_session, page_id)

            flash_messages = get_flash_messages(request)
            return TemplateResponse(
                "admin/pages/revisions.html",
                context={
                    "flash_messages": flash_messages,
                    "page": page,
                    "revisions": revisions,
                    **page_type_ctx,
                    **ctx,
                },
            )

        @post(
            f"/{plural}/{{page_id:uuid}}/revisions/{{revision_id:uuid}}/restore",
            guards=[auth_guard, OwnerOrPermission(perms["edit_own"], perms["manage"])],
        )
        async def restore_revision(
            self,
            request: Request,
            db_session: AsyncSession,
            page_id: UUID,
            revision_id: UUID,
        ) -> Redirect:
            page = await page_service.get_page_by_id(db_session, page_id)
            if redirect := _get_page_or_redirect(page, request):
                return redirect

            await check_page_access(
                db_session, request, page, perms["edit_own"], perms["manage"]
            )

            revision = await revision_service.get_revision(db_session, revision_id)
            if not revision or revision.page_id != page_id:
                flash_error(request, "Revision not found")
                return Redirect(path=f"{admin_base}/{page_id}/revisions")

            user_id = request.session.get(SESSION_USER_ID)
            await revision_service.restore_revision(
                db_session, page, revision, UUID(user_id) if user_id else None
            )

            flash_success(request, f"{label} restored to revision #{revision.revision_number}")
            return Redirect(path=f"{admin_base}/{page_id}/edit")

    # Give the class a unique name for Litestar's route registration
    _PageTypeController.__name__ = f"{label}AdminController"
    _PageTypeController.__qualname__ = f"{label}AdminController"

    return _PageTypeController



"""Association table linking pages to assets (many-to-many)."""

from sqlalchemy import Column, ForeignKey, Table

from skrift.db.base import Base

page_assets = Table(
    "page_assets",
    Base.metadata,
    Column("page_id", ForeignKey("pages.id", ondelete="CASCADE"), primary_key=True),
    Column("asset_id", ForeignKey("assets.id", ondelete="CASCADE"), primary_key=True),
)



"""Shared helpers for rendering public page responses."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from litestar import Request
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.controllers.helpers import get_user_context, resolve_theme
from skrift.db.services.asset_service import get_asset_url
from skrift.db.services.setting_service import (
    get_cached_site_base_url,
    get_cached_site_name,
)
from skrift.lib.seo import get_page_og_meta, get_page_seo_meta
from skrift.lib.storage import StorageManager


@dataclass
class PublicPageRenderContext:
    """Resolved data required to render a public page template."""

    asset_urls: dict[str, str]
    featured_image_url: str | None
    flash: Any
    og_meta: dict[str, Any]
    seo_meta: dict[str, Any]
    theme_name: str
    user_ctx: dict[str, Any]


def wants_markdown_response(request: Request) -> bool:
    """Return whether the client explicitly requested Markdown."""
    return "text/markdown" in request.headers.get("accept", "")


async def build_public_page_render_context(
    request: Request,
    db_session: AsyncSession,
    page: Any,
    *,
    include_asset_urls: bool = False,
) -> PublicPageRenderContext:
    """Resolve shared page context used by public controllers."""
    user_ctx = await get_user_context(request, db_session)
    flash = request.session.pop("flash", None)
    theme_name = await resolve_theme(request)

    asset_urls: dict[str, str] = {}
    featured_image_url = None
    storage: StorageManager | None = None

    if include_asset_urls and page.assets:
        storage = request.app.state.storage_manager
        asset_urls = {
            str(asset.id): await get_asset_url(storage, asset)
            for asset in page.assets
        }

    if page.featured_asset:
        featured_asset_key = str(page.featured_asset.id)
        featured_image_url = asset_urls.get(featured_asset_key)
        if featured_image_url is None:
            storage = storage or request.app.state.storage_manager
            featured_image_url = await get_asset_url(storage, page.featured_asset)

    site_name = get_cached_site_name()
    base_url = get_cached_site_base_url() or str(request.base_url).rstrip("/")
    seo_meta = await get_page_seo_meta(page, site_name, base_url)
    og_meta = await get_page_og_meta(
        page,
        site_name,
        base_url,
        featured_image_url=featured_image_url,
    )

    return PublicPageRenderContext(
        asset_urls=asset_urls,
        featured_image_url=featured_image_url,
        flash=flash,
        og_meta=og_meta,
        seo_meta=seo_meta,
        theme_name=theme_name,
        user_ctx=user_ctx,
    )



"""Add push_subscriptions table for Web Push notifications.

Revision ID: 4ef00221df53
Revises: s1t2u3v4w5x6
Create Date: 2026-03-08
"""

from alembic import op
import sqlalchemy as sa

revision = "4ef00221df53"
down_revision = "s1t2u3v4w5x6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "push_subscriptions",
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("user_id", sa.String(255), nullable=False),
        sa.Column("endpoint", sa.Text(), nullable=False),
        sa.Column("key_p256dh", sa.String(255), nullable=False),
        sa.Column("key_auth", sa.String(255), nullable=False),
        sa.Column("last_used_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()),
        sa.Column("sa_orm_sentinel", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("endpoint", name="uq_push_subscriptions_endpoint"),
    )
    op.create_index("ix_push_subscriptions_user_id", "push_subscriptions", ["user_id"])


def downgrade() -> None:
    op.drop_index("ix_push_subscriptions_user_id", table_name="push_subscriptions")
    op.drop_table("push_subscriptions")



"""Add assets table for file/media storage.

Revision ID: g9h0i1j2k3l4
Revises: f8a9b0c1d2e3
Create Date: 2026-02-27
"""

from alembic import op
import sqlalchemy as sa

revision = "g9h0i1j2k3l4"
down_revision = "f8a9b0c1d2e3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "assets",
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("key", sa.String(1024), nullable=False),
        sa.Column("store", sa.String(64), nullable=False, server_default="default"),
        sa.Column("content_hash", sa.String(64), nullable=False),
        sa.Column("filename", sa.String(255), nullable=False),
        sa.Column("content_type", sa.String(128), nullable=False),
        sa.Column("size", sa.Integer(), nullable=False),
        sa.Column("folder", sa.String(512), nullable=False, server_default=""),
        sa.Column("alt_text", sa.String(500), nullable=False, server_default=""),
        sa.Column("user_id", sa.Uuid(), nullable=True),
        sa.Column("sa_orm_sentinel", sa.Integer(), nullable=True),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.ForeignKeyConstraint(["user_id"], ["users.id"]),
    )
    op.create_index("ix_assets_key", "assets", ["key"])
    op.create_index("ix_assets_store", "assets", ["store"])
    op.create_index("ix_assets_content_hash", "assets", ["content_hash"])
    op.create_index("ix_assets_folder", "assets", ["folder"])
    op.create_index("ix_assets_user_id", "assets", ["user_id"])
    op.create_index("ix_asset_store_content_hash", "assets", ["store", "content_hash"])


def downgrade() -> None:
    op.drop_index("ix_asset_store_content_hash", table_name="assets")
    op.drop_index("ix_assets_user_id", table_name="assets")
    op.drop_index("ix_assets_folder", table_name="assets")
    op.drop_index("ix_assets_content_hash", table_name="assets")
    op.drop_index("ix_assets_store", table_name="assets")
    op.drop_index("ix_assets_key", table_name="assets")
    op.drop_table("assets")



"""OAuth2 scope definitions for the authorization server."""

from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class ScopeDefinition:
    """Definition of an OAuth2 scope with its associated claims."""

    name: str
    description: str
    claims: list[str] = field(default_factory=list)


# Registry of all scope definitions
SCOPE_DEFINITIONS: dict[str, ScopeDefinition] = {}


def register_scope(name: str, description: str, claims: list[str] | None = None) -> ScopeDefinition:
    """Register a scope definition.

    Args:
        name: The scope identifier (e.g., "openid", "profile")
        description: Human-readable description of what this scope grants
        claims: List of claim names included when this scope is granted

    Returns:
        The registered ScopeDefinition instance
    """
    scope = ScopeDefinition(name=name, description=description, claims=claims or [])
    SCOPE_DEFINITIONS[scope.name] = scope
    return scope


def get_scope_definition(name: str) -> ScopeDefinition | None:
    """Get a scope definition by name."""
    return SCOPE_DEFINITIONS.get(name)


# Built-in scopes
register_scope("openid", "Verify your identity", claims=["sub"])
register_scope("profile", "Access your name and picture", claims=["name", "picture"])
register_scope("email", "Access your email address", claims=["email"])



"""Asset storage service — upload, delete, list, URL resolution."""

from __future__ import annotations

import hashlib
from uuid import UUID

from sqlalchemy import and_, delete, func, select
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.db.models.asset import Asset
from skrift.db.models.page_asset import page_assets
from skrift.lib.hooks import (
    AFTER_ASSET_DELETE,
    AFTER_ASSET_UPLOAD,
    ASSET_UPLOAD_DATA,
    ASSET_UPLOAD_KEY,
    BEFORE_ASSET_DELETE,
    BEFORE_ASSET_UPLOAD,
    hooks,
)
from skrift.lib.storage.manager import StorageManager


class UploadTooLargeError(Exception):
    """Raised when an upload exceeds the configured size limit."""


async def upload_asset(
    db_session: AsyncSession,
    storage: StorageManager,
    filename: str,
    data: bytes,
    content_type: str,
    store: str | None = None,
    folder: str = "",
    user_id: UUID | None = None,
) -> Asset:
    """Upload a file and create an Asset record.

    Deduplicates by content hash — if the same bytes already exist in the
    store the backend ``put`` is skipped and only a new DB row is created.
    """
    store_name = store or storage.default_store
    backend = await storage.get(store_name)

    # Enforce size limit
    store_cfg = storage._config.stores.get(store_name)
    if store_cfg and len(data) > store_cfg.max_upload_size:
        raise UploadTooLargeError(
            f"File size {len(data)} exceeds limit {store_cfg.max_upload_size}"
        )

    # Apply data filter hook
    data = await hooks.apply_filters(ASSET_UPLOAD_DATA, data, filename=filename)

    # Compute content hash and build key
    content_hash = hashlib.sha256(data).hexdigest()
    key = content_hash

    # Apply key filter hook
    key = await hooks.apply_filters(ASSET_UPLOAD_KEY, key, filename=filename, content_hash=content_hash)

    # Fire before-upload action
    await hooks.do_action(BEFORE_ASSET_UPLOAD, filename=filename, key=key, store=store_name)

    # Dedup: skip backend write if identical content already stored
    existing = await db_session.execute(
        select(Asset).where(
            and_(Asset.store == store_name, Asset.content_hash == content_hash)
        ).limit(1)
    )
    if not existing.scalar_one_or_none():
        await backend.put(key, data, content_type)

    # Create Asset row
    asset = Asset(
        key=key,
        store=store_name,
        content_hash=content_hash,
        filename=filename,
        content_type=content_type,
        size=len(data),
        folder=folder,
        user_id=user_id,
    )
    db_session.add(asset)
    await db_session.commit()
    await db_session.refresh(asset)

    # Fire after-upload action
    await hooks.do_action(AFTER_ASSET_UPLOAD, asset)

    return asset


async def delete_asset(
    db_session: AsyncSession,
    storage: StorageManager,
    asset_id: UUID,
) -> bool:
    """Delete an asset. Removes backend file only when no other rows reference it."""
    result = await db_session.execute(select(Asset).where(Asset.id == asset_id))
    asset = result.scalar_one_or_none()
    if not asset:
        return False

    await hooks.do_action(BEFORE_ASSET_DELETE, asset)

    store_name = asset.store
    key = asset.key
    content_hash = asset.content_hash

    await db_session.delete(asset)
    await db_session.commit()

    # Check remaining references
    count_result = await db_session.execute(
        select(func.count()).select_from(Asset).where(
            and_(Asset.store == store_name, Asset.content_hash == content_hash)
        )
    )
    remaining = count_result.scalar() or 0

    if remaining == 0:
        backend = await storage.get(store_name)
        await backend.delete(key)

    await hooks.do_action(AFTER_ASSET_DELETE, asset_id=asset_id, key=key, store=store_name)

    return True


async def get_asset_url(storage: StorageManager, asset: Asset) -> str:
    """Return the public/signed URL for an asset."""
    backend = await storage.get(asset.store)
    return await backend.get_url(asset.key)


async def list_assets(
    db_session: AsyncSession,
    store: str | None = None,
    folder: str | None = None,
    content_type_prefix: str | None = None,
    limit: int = 50,
    offset: int = 0,
) -> list[Asset]:
    """List assets with optional filtering."""
    query = select(Asset)

    filters = []
    if store is not None:
        filters.append(Asset.store == store)
    if folder is not None:
        filters.append(Asset.folder == folder)
    if content_type_prefix is not None:
        filters.append(Asset.content_type.startswith(content_type_prefix))
    if filters:
        query = query.where(and_(*filters))

    query = query.order_by(Asset.created_at.desc())
    if offset:
        query = query.offset(offset)
    query = query.limit(limit)

    result = await db_session.execute(query)
    return list(result.scalars().all())


async def count_assets(
    db_session: AsyncSession,
    store: str | None = None,
    folder: str | None = None,
    content_type_prefix: str | None = None,
) -> int:
    """Count assets matching the given filters."""
    query = select(func.count()).select_from(Asset)

    filters = []
    if store is not None:
        filters.append(Asset.store == store)
    if folder is not None:
        filters.append(Asset.folder == folder)
    if content_type_prefix is not None:
        filters.append(Asset.content_type.startswith(content_type_prefix))
    if filters:
        query = query.where(and_(*filters))

    result = await db_session.execute(query)
    return result.scalar() or 0


async def get_page_asset_ids(
    db_session: AsyncSession,
    page_id: UUID,
) -> list[UUID]:
    """Return current asset IDs attached to a page."""
    result = await db_session.execute(
        select(page_assets.c.asset_id).where(page_assets.c.page_id == page_id)
    )
    return list(result.scalars().all())


async def sync_page_assets(
    db_session: AsyncSession,
    page_id: UUID,
    asset_ids: list[UUID],
) -> None:
    """Sync the page_assets rows so the page is linked to exactly `asset_ids`."""
    current = set(await get_page_asset_ids(db_session, page_id))
    desired = set(asset_ids)

    to_remove = current - desired
    to_add = desired - current

    if to_remove:
        await db_session.execute(
            delete(page_assets).where(
                and_(
                    page_assets.c.page_id == page_id,
                    page_assets.c.asset_id.in_(to_remove),
                )
            )
        )

    if to_add:
        await db_session.execute(
            page_assets.insert(),
            [{"page_id": page_id, "asset_id": aid} for aid in to_add],
        )

    if to_remove or to_add:
        await db_session.commit()



"""Storage backend protocol and common types."""

from __future__ import annotations

from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import Protocol, runtime_checkable


@dataclass
class StoredFile:
    """Metadata for a file stored in a backend."""

    key: str
    url: str
    content_type: str
    size: int
    content_hash: str


@runtime_checkable
class StorageBackend(Protocol):
    """Interface for pluggable asset storage backends."""

    async def put(self, key: str, data: bytes, content_type: str) -> StoredFile:
        """Store data under the given key."""
        ...

    async def get(self, key: str) -> bytes:
        """Retrieve the raw bytes for a key."""
        ...

    async def delete(self, key: str) -> None:
        """Remove a key from storage."""
        ...

    async def exists(self, key: str) -> bool:
        """Check whether a key exists in storage."""
        ...

    async def list_keys(self, prefix: str = "") -> AsyncIterator[str]:
        """Yield all keys matching the given prefix."""
        ...

    async def get_url(self, key: str) -> str:
        """Return a public or signed URL for the key."""
        ...



"""Media library admin controller."""

from __future__ import annotations

import logging
from typing import Annotated
from uuid import UUID

from litestar import Controller, Request, get, post
from litestar.datastructures import UploadFile
from litestar.enums import RequestEncodingType
from litestar.params import Body
from litestar.response import Redirect, Response
from litestar.response import Template as TemplateResponse
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.admin.helpers import get_admin_context
from skrift.auth.session_keys import SESSION_USER_ID
from skrift.admin.navigation import ADMIN_NAV_TAG
from skrift.auth.guards import Permission, auth_guard
from skrift.db.services.asset_service import (
    UploadTooLargeError,
    count_assets,
    delete_asset,
    get_asset_url,
    list_assets,
    upload_asset,
)
from skrift.lib.flash import flash_error, flash_success, get_flash_messages
from skrift.lib.storage import StorageManager

logger = logging.getLogger(__name__)


class MediaAdminController(Controller):
    """Controller for the admin media library."""

    path = "/admin"
    guards = [auth_guard]

    @get(
        "/media",
        tags=[ADMIN_NAV_TAG],
        guards=[auth_guard, Permission("manage-media")],
        opt={"label": "Media", "icon": "image", "order": 15},
    )
    async def media_library(
        self,
        request: Request,
        db_session: AsyncSession,
        page: int = 1,
        store: str | None = None,
    ) -> TemplateResponse:
        """Browse uploaded media assets."""
        ctx = await get_admin_context(request, db_session)

        per_page = 24
        offset = (page - 1) * per_page

        assets = await list_assets(db_session, store=store, limit=per_page, offset=offset)
        total = await count_assets(db_session, store=store)

        storage: StorageManager = request.app.state.storage_manager
        asset_urls = {
            str(asset.id): await get_asset_url(storage, asset)
            for asset in assets
        }

        total_pages = max(1, (total + per_page - 1) // per_page)

        flash_messages = get_flash_messages(request)
        return TemplateResponse(
            "admin/media/library.html",
            context={
                "flash_messages": flash_messages,
                "assets": assets,
                "asset_urls": asset_urls,
                "page": page,
                "total_pages": total_pages,
                "total": total,
                "store_names": storage.store_names,
                "current_store": store,
                **ctx,
            },
        )

    @post(
        "/media/upload",
        guards=[auth_guard, Permission("upload-media")],
    )
    async def upload_media(
        self,
        request: Request,
        db_session: AsyncSession,
        data: Annotated[UploadFile, Body(media_type=RequestEncodingType.MULTI_PART)],
    ) -> Redirect:
        """Handle file upload from the media library form."""
        storage: StorageManager = request.app.state.storage_manager
        user = getattr(request, "user", None)
        user_id = user.id if user else None

        content = await data.read()
        filename = data.filename or "untitled"
        content_type = data.content_type or "application/octet-stream"

        try:
            await upload_asset(
                db_session,
                storage,
                filename=filename,
                data=content,
                content_type=content_type,
                user_id=user_id,
            )
            flash_success(request, f"Uploaded {filename}")
        except UploadTooLargeError:
            flash_error(request, f"File too large: {filename}")
        except Exception:
            logger.exception("Admin media upload failed for '%s'", filename)
            flash_error(request, "Upload failed. Check the server logs and try again.")

        return Redirect(path="/admin/media")

    @post(
        "/media/upload-json",
        guards=[auth_guard, Permission("upload-media")],
    )
    async def upload_media_json(
        self,
        request: Request,
        db_session: AsyncSession,
        data: Annotated[UploadFile, Body(media_type=RequestEncodingType.MULTI_PART)],
    ) -> Response:
        """Handle file upload and return JSON (used by page editor fetch)."""
        storage: StorageManager = request.app.state.storage_manager
        uid = request.session.get(SESSION_USER_ID)
        user_id = UUID(uid) if uid else None

        content = await data.read()
        filename = data.filename or "untitled"
        content_type = data.content_type or "application/octet-stream"

        try:
            asset = await upload_asset(
                db_session,
                storage,
                filename=filename,
                data=content,
                content_type=content_type,
                user_id=user_id,
            )
        except UploadTooLargeError:
            return Response(
                content={"error": f"File too large: {filename}"},
                status_code=413,
                media_type="application/json",
            )
        except Exception:
            logger.exception("Admin media JSON upload failed for '%s'", filename)
            return Response(
                content={"error": "Upload failed. Check the server logs."},
                status_code=500,
                media_type="application/json",
            )

        url = await get_asset_url(storage, asset)
        return Response(
            content={
                "id": str(asset.id),
                "filename": asset.filename,
                "url": url,
                "content_type": asset.content_type,
                "size": asset.size,
            },
            status_code=201,
            media_type="application/json",
        )

    @post(
        "/media/{asset_id:uuid}/delete",
        guards=[auth_guard, Permission("manage-media")],
    )
    async def delete_media(
        self,
        request: Request,
        db_session: AsyncSession,
        asset_id: UUID,
    ) -> Redirect:
        """Delete an asset."""
        storage: StorageManager = request.app.state.storage_manager
        deleted = await delete_asset(db_session, storage, asset_id)

        if deleted:
            flash_success(request, "Asset deleted.")
        else:
            flash_error(request, "Asset not found.")

        return Redirect(path="/admin/media")

    @get(
        "/media/picker",
        guards=[auth_guard, Permission("upload-media")],
    )
    async def media_picker(
        self,
        request: Request,
        db_session: AsyncSession,
        page: int = 1,
    ) -> TemplateResponse:
        """Embeddable asset picker fragment for editors."""
        per_page = 20
        offset = (page - 1) * per_page
        assets = await list_assets(db_session, limit=per_page, offset=offset)
        total = await count_assets(db_session)

        storage: StorageManager = request.app.state.storage_manager
        asset_urls = {
            str(asset.id): await get_asset_url(storage, asset)
            for asset in assets
        }

        total_pages = max(1, (total + per_page - 1) // per_page)

        return TemplateResponse(
            "admin/media/picker.html",
            context={
                "assets": assets,
                "asset_urls": asset_urls,
                "page": page,
                "total_pages": total_pages,
            },
        )



"""Revoked token model for OAuth2 token revocation."""

from datetime import datetime

from sqlalchemy import String, DateTime
from sqlalchemy.orm import Mapped, mapped_column

from skrift.db.base import Base


class RevokedToken(Base):
    """A revoked OAuth2 token tracked by its JTI."""

    __tablename__ = "revoked_tokens"

    jti: Mapped[str] = mapped_column(String(64), unique=True, index=True, nullable=False)
    token_type: Mapped[str] = mapped_column(String(20), nullable=False)
    expires_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), index=True, nullable=False)



"""S3-compatible storage backend (requires ``pip install skrift[s3]``)."""

from __future__ import annotations

import hashlib
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

try:
    import aioboto3
except ImportError as exc:
    raise ImportError(
        "S3 storage backend requires aioboto3. Install it with: pip install skrift[s3]"
    ) from exc

from skrift.lib.storage.base import StoredFile

if TYPE_CHECKING:
    from skrift.config import S3Config


class S3StorageBackend:
    """Store files in an S3-compatible bucket."""

    def __init__(self, config: S3Config) -> None:
        self._config = config
        self._session = aioboto3.Session()

    def _client_kwargs(self) -> dict:
        kwargs: dict = {
            "region_name": self._config.region,
        }
        if self._config.endpoint_url:
            kwargs["endpoint_url"] = self._config.endpoint_url
        if self._config.access_key_id:
            kwargs["aws_access_key_id"] = self._config.access_key_id
        if self._config.secret_access_key:
            kwargs["aws_secret_access_key"] = self._config.secret_access_key
        return kwargs

    def _full_key(self, key: str) -> str:
        if self._config.prefix:
            return f"{self._config.prefix.rstrip('/')}/{key}"
        return key

    async def put(self, key: str, data: bytes, content_type: str) -> StoredFile:
        full_key = self._full_key(key)
        put_kwargs: dict = {
            "Bucket": self._config.bucket,
            "Key": full_key,
            "Body": data,
            "ContentType": content_type,
        }
        if self._config.acl:
            put_kwargs["ACL"] = self._config.acl

        async with self._session.client("s3", **self._client_kwargs()) as s3:
            await s3.put_object(**put_kwargs)

        return StoredFile(
            key=key,
            url=await self.get_url(key),
            content_type=content_type,
            size=len(data),
            content_hash=hashlib.sha256(data).hexdigest(),
        )

    async def get(self, key: str) -> bytes:
        full_key = self._full_key(key)
        async with self._session.client("s3", **self._client_kwargs()) as s3:
            response = await s3.get_object(Bucket=self._config.bucket, Key=full_key)
            return await response["Body"].read()

    async def delete(self, key: str) -> None:
        full_key = self._full_key(key)
        async with self._session.client("s3", **self._client_kwargs()) as s3:
            await s3.delete_object(Bucket=self._config.bucket, Key=full_key)

    async def exists(self, key: str) -> bool:
        full_key = self._full_key(key)
        async with self._session.client("s3", **self._client_kwargs()) as s3:
            try:
                await s3.head_object(Bucket=self._config.bucket, Key=full_key)
                return True
            except s3.exceptions.ClientError:
                return False

    async def list_keys(self, prefix: str = "") -> AsyncIterator[str]:
        full_prefix = self._full_key(prefix)
        async with self._session.client("s3", **self._client_kwargs()) as s3:
            paginator = s3.get_paginator("list_objects_v2")
            async for page in paginator.paginate(
                Bucket=self._config.bucket, Prefix=full_prefix
            ):
                for obj in page.get("Contents", []):
                    key = obj["Key"]
                    # Strip prefix to return relative keys
                    if self._config.prefix and key.startswith(self._config.prefix):
                        key = key[len(self._config.prefix.rstrip("/")) + 1:]
                    yield key

    async def get_url(self, key: str) -> str:
        full_key = self._full_key(key)

        # CDN / custom public URL
        if self._config.public_url:
            base = self._config.public_url.rstrip("/")
            return f"{base}/{full_key}"

        # Public-read bucket: standard S3 URL
        if self._config.acl == "public-read":
            if self._config.endpoint_url:
                base = self._config.endpoint_url.rstrip("/")
                return f"{base}/{self._config.bucket}/{full_key}"
            return f"https://{self._config.bucket}.s3.{self._config.region}.amazonaws.com/{full_key}"

        # Private: generate presigned URL
        async with self._session.client("s3", **self._client_kwargs()) as s3:
            return await s3.generate_presigned_url(
                "get_object",
                Params={"Bucket": self._config.bucket, "Key": full_key},
                ExpiresIn=self._config.presign_ttl,
            )

    async def close(self) -> None:
        """No persistent resources to clean up."""



"""Dynamic controller factory for public-facing page type routes."""

from __future__ import annotations

from pathlib import Path

from litestar import Controller, Request, get
from litestar.exceptions import NotFoundException
from litestar.response import Response, Template as TemplateResponse
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.auth.session_keys import SESSION_USER_ID
from skrift.config import PageTypeConfig
from skrift.controllers.helpers import get_user_context, resolve_theme
from skrift.controllers.page_rendering import (
    build_public_page_render_context,
    wants_markdown_response,
)
from skrift.db.services import page_service
from skrift.lib.template import Template

TEMPLATE_DIR = Path(__file__).parent.parent.parent / "templates"


def create_public_page_type_controller(
    page_type: PageTypeConfig,
    for_subdomain: bool = False,
) -> type[Controller]:
    """Create a public Controller subclass for a specific page type.

    Primary domain mode (for_subdomain=False):
        Routes under /{type_name}, e.g. /post/hello-world

    Subdomain mode (for_subdomain=True):
        Routes under /, e.g. blog.example.com/hello-world
    """
    type_name = page_type.name
    plural = page_type.plural
    label = type_name.title()
    base_path = "/" if for_subdomain else f"/{type_name}"

    page_type_ctx = {
        "page_type_name": type_name,
        "page_type_plural": plural,
    }

    class _PublicPageTypeController(Controller):
        path = base_path

        @get("/")
        async def list_pages(
            self, request: Request, db_session: AsyncSession
        ) -> TemplateResponse:
            user_ctx = await get_user_context(request, db_session)
            flash = request.session.pop("flash", None)
            theme_name = await resolve_theme(request)

            pages = await page_service.list_pages(
                db_session,
                published_only=not request.session.get(SESSION_USER_ID),
                page_type=type_name,
                order_by="published",
            )

            if for_subdomain:
                template = Template("index", context={"pages": pages})
            else:
                template = Template(
                    "archive", type_name,
                    context={"pages": pages},
                )

            return template.render(
                TEMPLATE_DIR,
                theme_name=theme_name,
                flash=flash,
                **page_type_ctx,
                **user_ctx,
            )

        @get("/{slug:str}")
        async def view_page(
            self, request: Request, db_session: AsyncSession, slug: str
        ) -> TemplateResponse | Response:
            page = await page_service.get_page_by_slug(
                db_session,
                slug,
                published_only=not request.session.get(SESSION_USER_ID),
                page_type=type_name,
            )
            if not page:
                raise NotFoundException(f"{label} '{slug}' not found")

            if wants_markdown_response(request):
                return Response(
                    content=page.content,
                    media_type="text/markdown",
                )

            render_ctx = await build_public_page_render_context(
                request,
                db_session,
                page,
                include_asset_urls=True,
            )

            template = Template(
                type_name, slug,
                context={
                    "page": page,
                    "seo_meta": render_ctx.seo_meta,
                    "og_meta": render_ctx.og_meta,
                    "featured_image_url": render_ctx.featured_image_url,
                    "asset_urls": render_ctx.asset_urls,
                },
            )
            return template.render(
                TEMPLATE_DIR,
                theme_name=render_ctx.theme_name,
                flash=render_ctx.flash,
                **page_type_ctx,
                **render_ctx.user_ctx,
            )

    suffix = "Subdomain" if for_subdomain else "Public"
    _PublicPageTypeController.__name__ = f"{label}{suffix}Controller"
    _PublicPageTypeController.__qualname__ = f"{label}{suffix}Controller"

    return _PublicPageTypeController



from skrift.db.models.asset import Asset
from skrift.db.models.notification import DismissedNotification, StoredNotification
from skrift.db.models.oauth2_client import OAuth2Client
from skrift.db.models.oauth_account import OAuthAccount
from skrift.db.models.page import Page
from skrift.db.models.page_asset import page_assets
from skrift.db.models.page_revision import PageRevision
from skrift.db.models.push_subscription import PushSubscription
from skrift.db.models.revoked_token import RevokedToken
from skrift.db.models.role import Role, RolePermission, user_roles
from skrift.db.models.setting import Setting
from skrift.db.models.user import User

__all__ = ["Asset", "DismissedNotification", "OAuth2Client", "OAuthAccount", "Page", "PageRevision", "PushSubscription", "RevokedToken", "Role", "RolePermission", "Setting", "StoredNotification", "User", "page_assets", "user_roles"]



"""Storage manager — registry of named storage backends."""

from __future__ import annotations

import importlib
from typing import TYPE_CHECKING

from skrift.lib.storage.local import LocalStorageBackend

if TYPE_CHECKING:
    from pathlib import Path

    from skrift.config import StorageConfig, StoreConfig
    from skrift.lib.storage.base import StorageBackend


class StorageManager:
    """Registry that lazily creates and caches storage backends by name."""

    def __init__(self, config: StorageConfig) -> None:
        self._config = config
        self._backends: dict[str, StorageBackend] = {}

    @property
    def default_store(self) -> str:
        return self._config.default

    @property
    def store_names(self) -> list[str]:
        return list(self._config.stores.keys())

    async def get(self, name: str | None = None) -> StorageBackend:
        """Return the backend for *name*, creating it on first access."""
        name = name or self._config.default
        if name not in self._backends:
            store_cfg = self._config.stores.get(name)
            if store_cfg is None:
                raise KeyError(f"Unknown storage store: {name!r}")
            self._backends[name] = create_storage_backend(store_cfg, store_name=name)
        return self._backends[name]

    async def close(self) -> None:
        """Release resources held by backends."""
        for backend in self._backends.values():
            close = getattr(backend, "close", None)
            if close is not None:
                await close()
        self._backends.clear()


def create_storage_backend(config: StoreConfig, store_name: str = "default") -> StorageBackend:
    """Instantiate a storage backend from configuration."""
    from pathlib import Path

    backend_type = config.backend

    if backend_type == "local":
        return LocalStorageBackend(
            base_path=Path(config.local_path),
            store_name=store_name,
        )

    if backend_type == "s3":
        from skrift.lib.storage.s3 import S3StorageBackend

        return S3StorageBackend(config.s3)

    # Dynamic import: "module:ClassName"
    if ":" in backend_type:
        parts = backend_type.split(":")
        if len(parts) != 2:
            raise ValueError(
                f"Invalid backend spec '{backend_type}': must contain exactly one colon"
            )
        module_path, class_name = parts
        module = importlib.import_module(module_path)
        cls = getattr(module, class_name)
        return cls(config)

    raise ValueError(
        f"Unknown storage backend '{backend_type}'. "
        "Use 'local', 's3', or 'module:ClassName'."
    )



"""Alembic environment configuration for async SQLAlchemy migrations."""

import asyncio
import logging
from logging.config import fileConfig

import sqlalchemy as sa
from alembic import context
from sqlalchemy import pool
from sqlalchemy.engine import Connection
from sqlalchemy.ext.asyncio import async_engine_from_config

from skrift.config import get_settings
from skrift.db.base import Base

logger = logging.getLogger("alembic.env")

# Import all models to ensure they're registered with Base.metadata
from skrift.db.models.asset import Asset  # noqa: F401
from skrift.db.models.notification import DismissedNotification, NotificationSubscription, StoredNotification  # noqa: F401
from skrift.db.models.oauth_account import OAuthAccount  # noqa: F401
from skrift.db.models.page import Page  # noqa: F401
from skrift.db.models.page_revision import PageRevision  # noqa: F401
from skrift.db.models.role import Role, RolePermission  # noqa: F401
from skrift.db.models.setting import Setting  # noqa: F401
from skrift.db.models.user import User  # noqa: F401

# Dynamically import user model modules from app.yaml
import importlib
import os
import sys

from skrift.config import load_model_modules

# Ensure project cwd is on sys.path for user module imports
_cwd = os.getcwd()
if _cwd not in sys.path:
    sys.path.insert(0, _cwd)

for _module_path in load_model_modules():
    try:
        importlib.import_module(_module_path)
    except ImportError:
        pass  # Graceful fallback if models can't be loaded

# Alembic Config object
config = context.config

# Set up logging from alembic.ini
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# Target metadata for 'autogenerate' support
target_metadata = Base.metadata


def get_url() -> str:
    """Get database URL from settings, app.yaml, or alembic.ini.

    Tries in order:
    1. Full settings (requires SECRET_KEY and all env vars)
    2. Direct app.yaml read (works during setup when SECRET_KEY isn't set)
    3. alembic.ini fallback (last resort)
    """
    try:
        settings = get_settings()
        return settings.db.url
    except BaseException:
        pass

    # Settings failed (e.g. SECRET_KEY not set during setup).
    # Read the database URL directly from app.yaml to avoid running
    # migrations against the wrong database.
    try:
        from skrift.setup.state import get_database_url_from_yaml
        url = get_database_url_from_yaml()
        if url:
            logger.info("Using database URL from app.yaml (settings unavailable)")
            return url
    except Exception:
        logger.debug("Could not resolve database URL from app.yaml", exc_info=True)

    fallback = config.get_main_option("sqlalchemy.url", "")
    logger.warning("Falling back to alembic.ini database URL: %s", fallback)
    return fallback


def get_schema() -> str | None:
    """Get database schema from settings or app.yaml."""
    try:
        settings = get_settings()
        return settings.db.db_schema
    except Exception:
        logger.debug("Could not resolve schema from settings", exc_info=True)

    # Fall back to reading schema directly from app.yaml
    try:
        from skrift.setup.state import get_database_schema_from_yaml
        return get_database_schema_from_yaml()
    except Exception:
        logger.debug("Could not resolve schema from app.yaml", exc_info=True)
        return None


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL and not an Engine.
    Calls to context.execute() emit the SQL to the script output.
    """
    url = get_url()
    schema = get_schema()

    if schema:
        Base.metadata.schema = schema

    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        include_schemas=bool(schema),
        version_table_schema=schema,
    )

    with context.begin_transaction():
        if schema:
            context.execute(f"SET search_path TO {schema}")
        context.run_migrations()


def do_run_migrations(connection: Connection) -> None:
    """Run migrations within a connection context."""
    schema = get_schema()

    if schema:
        Base.metadata.schema = schema

    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        include_schemas=bool(schema),
        version_table_schema=schema,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations() -> None:
    """Run migrations in 'online' mode with async engine."""
    configuration = config.get_section(config.config_ini_section, {})
    configuration["sqlalchemy.url"] = get_url()

    connectable = async_engine_from_config(
        configuration,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    async with connectable.connect() as connection:
        schema = get_schema()
        if schema:
            await connection.execute(sa.text(f"SET search_path TO {schema}"))
        await connection.run_sync(do_run_migrations)
        await connection.commit()

    await connectable.dispose()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode."""
    asyncio.run(run_async_migrations())


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Asset model for file/media storage."""

from __future__ import annotations

from uuid import UUID

from sqlalchemy import ForeignKey, Index, Integer, String
from sqlalchemy.orm import Mapped, mapped_column

from skrift.db.base import Base


class Asset(Base):
    """A stored file or media asset."""

    __tablename__ = "assets"
    __table_args__ = (
        Index("ix_asset_store_content_hash", "store", "content_hash"),
    )

    key: Mapped[str] = mapped_column(String(1024), nullable=False, index=True)
    store: Mapped[str] = mapped_column(String(64), nullable=False, default="default", index=True)
    content_hash: Mapped[str] = mapped_column(String(64), nullable=False, index=True)
    filename: Mapped[str] = mapped_column(String(255), nullable=False)
    content_type: Mapped[str] = mapped_column(String(128), nullable=False)
    size: Mapped[int] = mapped_column(Integer, nullable=False)
    folder: Mapped[str] = mapped_column(String(512), nullable=False, default="", index=True)
    alt_text: Mapped[str] = mapped_column(String(500), nullable=False, default="")
    user_id: Mapped[UUID | None] = mapped_column(
        ForeignKey("users.id"), nullable=True, index=True
    )



"""OAuth2 authorization server database service."""

import secrets
from datetime import datetime
from uuid import UUID

from sqlalchemy import select, delete
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.db.models.oauth2_client import OAuth2Client
from skrift.db.models.revoked_token import RevokedToken
from skrift.lib.hooks import hooks


async def get_client_by_client_id(db_session: AsyncSession, client_id: str) -> OAuth2Client | None:
    """Look up an active OAuth2 client by its client_id string."""
    result = await db_session.execute(
        select(OAuth2Client).where(
            OAuth2Client.client_id == client_id,
            OAuth2Client.is_active == True,  # noqa: E712
        )
    )
    return result.scalar_one_or_none()


async def list_clients(db_session: AsyncSession) -> list[OAuth2Client]:
    """List all OAuth2 clients."""
    result = await db_session.execute(
        select(OAuth2Client).order_by(OAuth2Client.created_at.desc())
    )
    return list(result.scalars().all())


async def create_client(
    db_session: AsyncSession,
    display_name: str,
    redirect_uris: list[str],
    allowed_scopes: list[str],
) -> OAuth2Client:
    """Create a new OAuth2 client with auto-generated credentials."""
    client = OAuth2Client(
        client_id=secrets.token_urlsafe(24),
        client_secret=secrets.token_urlsafe(48),
        display_name=display_name,
        redirect_uris="\n".join(redirect_uris),
        allowed_scopes="\n".join(allowed_scopes),
    )
    db_session.add(client)
    await db_session.commit()
    await hooks.do_action("after_oauth2_client_created", client)
    return client


async def update_client(
    db_session: AsyncSession,
    client: OAuth2Client,
    display_name: str | None = None,
    redirect_uris: list[str] | None = None,
    allowed_scopes: list[str] | None = None,
    is_active: bool | None = None,
) -> OAuth2Client:
    """Update an existing OAuth2 client."""
    if display_name is not None:
        client.display_name = display_name
    if redirect_uris is not None:
        client.redirect_uris = "\n".join(redirect_uris)
    if allowed_scopes is not None:
        client.allowed_scopes = "\n".join(allowed_scopes)
    if is_active is not None:
        client.is_active = is_active
    await db_session.commit()
    await hooks.do_action("after_oauth2_client_updated", client)
    return client


async def delete_client(db_session: AsyncSession, client_id: UUID) -> None:
    """Delete an OAuth2 client by its database UUID."""
    result = await db_session.execute(
        select(OAuth2Client).where(OAuth2Client.id == client_id)
    )
    client = result.scalar_one_or_none()
    if client:
        await hooks.do_action("before_oauth2_client_deleted", client)
        await db_session.delete(client)
        await db_session.commit()
        await hooks.do_action("after_oauth2_client_deleted", client_id)


async def regenerate_client_secret(db_session: AsyncSession, client: OAuth2Client) -> str:
    """Regenerate a client's secret and return the new value."""
    new_secret = secrets.token_urlsafe(48)
    client.client_secret = new_secret
    await db_session.commit()
    await hooks.do_action("after_oauth2_client_secret_regenerated", client)
    return new_secret


async def revoke_token(
    db_session: AsyncSession,
    jti: str,
    token_type: str,
    expires_at: datetime,
) -> None:
    """Record a token revocation."""
    revoked = RevokedToken(jti=jti, token_type=token_type, expires_at=expires_at)
    db_session.add(revoked)
    await db_session.commit()
    await hooks.do_action("after_token_revoked", jti, token_type)


async def is_token_revoked(db_session: AsyncSession, jti: str) -> bool:
    """Check if a token has been revoked."""
    result = await db_session.execute(
        select(RevokedToken.id).where(RevokedToken.jti == jti)
    )
    return result.scalar_one_or_none() is not None


async def cleanup_expired_revocations(db_session: AsyncSession) -> int:
    """Delete revocation records for tokens that have already expired."""
    now = datetime.now(tz=datetime.now().astimezone().tzinfo)
    result = await db_session.execute(
        delete(RevokedToken).where(RevokedToken.expires_at < now)
    )
    await db_session.commit()
    return result.rowcount



"""Web Push notification support with VAPID authentication.

Provides three layers:

1. Low-level ``send_push()`` — sends to all subscriptions for a user.
2. VAPID key management — auto-generates on first use, cached in memory.
3. Unified ``notify()`` — SSE for connected clients, Web Push fallback
   for disconnected users. Single dispatch point for app code.

Usage:
    from skrift.lib.push import send_push, get_vapid_public_key, notify

    # Get the public key for frontend subscription
    public_key = await get_vapid_public_key(db_session)

    # Send a push notification to a user
    await send_push(
        db_session,
        user_id="some-uuid",
        title="New message",
        body="Claude needs your input",
        url="/chat/session-abc",
    )

    # Unified: SSE when connected, push when not
    await notify(
        db_session,
        user_id="some-uuid",
        event="new_message",
        data={"title": "New message", "body": "Claude is waiting"},
        push_fallback=True,
    )
"""

import base64
import json
import logging
from datetime import datetime, timezone
from uuid import UUID

from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession

logger = logging.getLogger(__name__)

# VAPID key setting keys
VAPID_PRIVATE_KEY = "webpush:vapid_private"
VAPID_PUBLIC_KEY = "webpush:vapid_public"

# In-memory cache for VAPID keys (loaded once from DB)
_vapid_private_key: str | None = None
_vapid_public_key: str | None = None


def _generate_vapid_keys() -> tuple[str, str]:
    """Generate a new VAPID keypair.

    Returns:
        Tuple of (private_key_b64url, public_key_b64url) in unpadded base64url format.
    """
    from cryptography.hazmat.primitives.asymmetric.ec import SECP256R1, generate_private_key
    from cryptography.hazmat.primitives.serialization import Encoding, NoEncryption, PrivateFormat, PublicFormat

    private_key = generate_private_key(SECP256R1())

    # DER-encoded private key (PKCS8) for pywebpush compatibility
    private_bytes = private_key.private_bytes(
        Encoding.DER, PrivateFormat.PKCS8, NoEncryption()
    )
    private_b64 = base64.urlsafe_b64encode(private_bytes).decode().rstrip("=")

    # Uncompressed public key point (65 bytes) for the browser
    public_bytes = private_key.public_key().public_bytes(
        Encoding.X962, PublicFormat.UncompressedPoint
    )
    public_b64 = base64.urlsafe_b64encode(public_bytes).decode().rstrip("=")

    return private_b64, public_b64


async def _ensure_vapid_keys(db_session: AsyncSession) -> tuple[str, str]:
    """Ensure VAPID keys exist, generating them on first use.

    Returns:
        Tuple of (private_key_b64url, public_key_b64url).
    """
    global _vapid_private_key, _vapid_public_key

    if _vapid_private_key and _vapid_public_key:
        return _vapid_private_key, _vapid_public_key

    from skrift.db.services.setting_service import get_setting, set_setting

    private = await get_setting(db_session, VAPID_PRIVATE_KEY)
    public = await get_setting(db_session, VAPID_PUBLIC_KEY)

    if private and public:
        _vapid_private_key = private
        _vapid_public_key = public
        return private, public

    # Generate new keypair
    private, public = _generate_vapid_keys()
    await set_setting(db_session, VAPID_PRIVATE_KEY, private)
    await set_setting(db_session, VAPID_PUBLIC_KEY, public)

    _vapid_private_key = private
    _vapid_public_key = public
    logger.info("Generated new VAPID keypair for Web Push")
    return private, public


async def get_vapid_public_key(db_session: AsyncSession) -> str:
    """Get the VAPID public key (base64url-encoded, unpadded).

    Generates a keypair on first use.
    """
    _, public = await _ensure_vapid_keys(db_session)
    return public


async def save_subscription(
    db_session: AsyncSession,
    user_id: str,
    endpoint: str,
    key_p256dh: str,
    key_auth: str,
) -> None:
    """Save or update a push subscription for a user."""
    from skrift.db.models.push_subscription import PushSubscription

    # Check if subscription already exists (by endpoint)
    result = await db_session.execute(
        select(PushSubscription).where(PushSubscription.endpoint == endpoint)
    )
    existing = result.scalar_one_or_none()

    if existing:
        existing.user_id = user_id
        existing.key_p256dh = key_p256dh
        existing.key_auth = key_auth
    else:
        sub = PushSubscription(
            user_id=user_id,
            endpoint=endpoint,
            key_p256dh=key_p256dh,
            key_auth=key_auth,
        )
        db_session.add(sub)

    await db_session.commit()


async def remove_subscription(db_session: AsyncSession, endpoint: str) -> bool:
    """Remove a push subscription by endpoint. Returns True if found."""
    from skrift.db.models.push_subscription import PushSubscription

    result = await db_session.execute(
        select(PushSubscription).where(PushSubscription.endpoint == endpoint)
    )
    sub = result.scalar_one_or_none()
    if not sub:
        return False

    await db_session.delete(sub)
    await db_session.commit()
    return True


async def send_push(
    db_session: AsyncSession,
    user_id: str,
    title: str,
    body: str,
    url: str | None = None,
    tag: str | None = None,
) -> int:
    """Send a Web Push notification to all of a user's subscribed browsers.

    Args:
        db_session: Database session
        user_id: Target user ID
        title: Notification title
        body: Notification body text
        url: Optional URL to open when notification is clicked
        tag: Optional tag for notification grouping/replacement

    Returns:
        Number of notifications successfully sent.
    """
    from pywebpush import WebPushException, webpush

    from skrift.db.models.push_subscription import PushSubscription

    private_key, _ = await _ensure_vapid_keys(db_session)

    # Fetch all subscriptions for this user
    result = await db_session.execute(
        select(PushSubscription).where(PushSubscription.user_id == user_id)
    )
    subscriptions = list(result.scalars().all())

    if not subscriptions:
        return 0

    # Build the push payload
    payload = {"title": title, "body": body}
    if url:
        payload["url"] = url
    if tag:
        payload["tag"] = tag
    data = json.dumps(payload)

    # Add padding back for pywebpush
    padded_key = private_key + "=" * (4 - len(private_key) % 4) if len(private_key) % 4 else private_key

    sent = 0
    expired_ids: list[UUID] = []

    for sub in subscriptions:
        subscription_info = {
            "endpoint": sub.endpoint,
            "keys": {
                "p256dh": sub.key_p256dh,
                "auth": sub.key_auth,
            },
        }

        try:
            webpush(
                subscription_info=subscription_info,
                data=data,
                vapid_private_key=padded_key,
                vapid_claims={"sub": "mailto:push@skrift.dev"},
            )
            sub.last_used_at = datetime.now(timezone.utc)
            sent += 1
        except WebPushException as e:
            if e.response and e.response.status_code in (404, 410):
                # Subscription expired or unsubscribed — clean up
                expired_ids.append(sub.id)
                logger.debug("Removing expired push subscription %s", sub.endpoint)
            else:
                logger.warning("Web Push failed for %s: %s", sub.endpoint, e)
        except Exception:
            logger.exception("Unexpected error sending push to %s", sub.endpoint)

    # Clean up expired subscriptions
    if expired_ids:
        await db_session.execute(
            delete(PushSubscription).where(PushSubscription.id.in_(expired_ids))
        )

    await db_session.commit()
    return sent


async def notify(
    db_session: AsyncSession,
    user_id: str,
    event: str,
    data: dict | None = None,
    *,
    push_fallback: bool = True,
    push_title: str | None = None,
    push_body: str | None = None,
    push_url: str | None = None,
    push_tag: str | None = None,
    group: str | None = None,
) -> None:
    """Unified notification dispatch: SSE for connected clients, Web Push for disconnected.

    Always sends via SSE. If ``push_fallback=True`` and the user has no active
    SSE connections, also sends a Web Push notification.

    Args:
        db_session: Database session
        user_id: Target user ID
        event: Notification event type (e.g. "new_message")
        data: Payload dict for the SSE notification
        push_fallback: If True, send push when no SSE connection is active
        push_title: Title for push notification (falls back to data["title"])
        push_body: Body for push notification (falls back to data["body"])
        push_url: URL to open on click (falls back to data["url"])
        push_tag: Tag for push grouping (falls back to event type)
        group: Notification group key for SSE
    """
    from skrift.lib.notifications import NotificationMode, notifications, notify_user

    payload = data or {}

    # Always send via SSE
    await notify_user(
        user_id,
        event,
        mode=NotificationMode.TIMESERIES,
        group=group,
        **payload,
    )

    # Check push_notify from payload: True=always, False=never, None=auto (fallback)
    push_notify = payload.get("push_notify")
    if push_notify is False or not push_fallback:
        return

    should_push = push_notify is True
    if not should_push:
        # Auto mode: only send push if no active SSE connection
        user_key = f"user:{user_id}"
        has_sse = notifications._registry.has_listeners(user_key)

        if not has_sse:
            children = notifications._registry._subscribers.get(user_key, set())
            has_sse = any(
                notifications._registry.has_listeners(child)
                for child in children
            )

        should_push = not has_sse

    if should_push:
            title = push_title or payload.get("title", event)
            body = push_body or payload.get("body", "")
            url = push_url or payload.get("url")
            tag = push_tag or event

            try:
                await send_push(
                    db_session,
                    user_id=user_id,
                    title=title,
                    body=body,
                    url=url,
                    tag=tag,
                )
            except Exception:
                logger.exception("Push fallback failed for user %s", user_id)


def setup_push_hook(session_maker) -> None:
    """Register a NOTIFICATION_SENT hook that triggers push fallback.

    Call this during app startup to enable automatic push notifications
    for users without active SSE connections.

    Args:
        session_maker: Async session factory (e.g. db_config.get_session)
    """
    from skrift.lib.hooks import NOTIFICATION_SENT, hooks
    from skrift.lib.notifications import Notification

    async def _push_on_notification(notification: Notification, scope: str, scope_id: str | None) -> None:
        """Hook: send push notification when user has no active SSE connection."""
        if scope != "user" or not scope_id:
            return

        # Check push_notify payload flag: True=always, False=never, None=auto
        push_notify = notification.payload.get("push_notify")
        if push_notify is False:
            return

        if push_notify is not True:
            # Auto mode: only send push if no active SSE connection
            from skrift.lib.notifications import notifications

            user_key = f"user:{scope_id}"
            has_sse = notifications._registry.has_listeners(user_key)
            if not has_sse:
                children = notifications._registry._subscribers.get(user_key, set())
                has_sse = any(
                    notifications._registry.has_listeners(child)
                    for child in children
                )

            if has_sse:
                return

        # No SSE — send push
        title = notification.payload.get("title", notification.type)
        body = notification.payload.get("body", "")
        url = notification.payload.get("url")
        tag = notification.payload.get("tag") or notification.group or notification.type

        try:
            async with session_maker() as db_session:
                await send_push(
                    db_session,
                    user_id=scope_id,
                    title=title,
                    body=body,
                    url=url,
                    tag=tag,
                )
        except Exception:
            logger.exception("Push hook failed for user %s", scope_id)

    hooks.add_action(NOTIFICATION_SENT, _push_on_notification, priority=50)



from pathlib import Path

from litestar import Controller, Request, get
from litestar.exceptions import NotFoundException
from litestar.response import Response, Template as TemplateResponse
from sqlalchemy.ext.asyncio import AsyncSession

from skrift.auth.session_keys import SESSION_USER_ID
from skrift.controllers.helpers import get_user_context
from skrift.controllers.page_rendering import (
    build_public_page_render_context,
    wants_markdown_response,
)
from skrift.db.services import page_service
from skrift.lib.template import Template

TEMPLATE_DIR = Path(__file__).parent.parent.parent / "templates"


class WebController(Controller):
    path = "/"

    @get("/")
    async def index(
        self, request: "Request", db_session: AsyncSession
    ) -> TemplateResponse:
        """Home page."""
        user_ctx = await get_user_context(request, db_session)
        flash = request.session.pop("flash", None)

        return TemplateResponse(
            "index.html",
            context={"flash": flash, **user_ctx},
        )

    @get("/{path:path}")
    async def view_page(
        self, request: "Request", db_session: AsyncSession, path: str
    ) -> TemplateResponse | Response:
        """View a page by path with WP-like template resolution."""
        # Split path into slugs (e.g., "services/web" -> ["services", "web"])
        slugs = [s for s in path.split("/") if s]

        # Use the full path as the slug for database lookup
        page_slug = "/".join(slugs)

        # Fetch page from database
        page = await page_service.get_page_by_slug(
            db_session, page_slug,
            published_only=not request.session.get(SESSION_USER_ID),
            page_type="page",
        )
        if not page:
            raise NotFoundException(f"Page '{path}' not found")

        if wants_markdown_response(request):
            return Response(
                content=page.content,
                media_type="text/markdown",
            )

        render_ctx = await build_public_page_render_context(request, db_session, page)

        template = Template(
            "page", *slugs,
            context={
                "path": path,
                "slugs": slugs,
                "page": page,
                "seo_meta": render_ctx.seo_meta,
                "og_meta": render_ctx.og_meta,
                "featured_image_url": render_ctx.featured_image_url,
            }
        )
        return template.render(
            TEMPLATE_DIR,
            theme_name=render_ctx.theme_name,
            flash=render_ctx.flash,
            **render_ctx.user_ctx,
        )



"""OAuth2 client model for the authorization server."""

from sqlalchemy import String, Text, Boolean
from sqlalchemy.orm import Mapped, mapped_column

from skrift.db.base import Base


class OAuth2Client(Base):
    """A registered OAuth2 client application."""

    __tablename__ = "oauth2_clients"

    client_id: Mapped[str] = mapped_column(String(255), unique=True, index=True, nullable=False)
    client_secret: Mapped[str] = mapped_column(String(255), default="", nullable=False)
    display_name: Mapped[str] = mapped_column(String(255), nullable=False)
    redirect_uris: Mapped[str] = mapped_column(Text, default="", nullable=False)
    allowed_scopes: Mapped[str] = mapped_column(Text, default="", nullable=False)
    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)

    @property
    def redirect_uri_list(self) -> list[str]:
        """Parse newline-delimited redirect_uris into a list."""
        return [uri.strip() for uri in self.redirect_uris.split("\n") if uri.strip()]

    @property
    def allowed_scope_list(self) -> list[str]:
        """Parse newline-delimited allowed_scopes into a list."""
        return [s.strip() for s in self.allowed_scopes.split("\n") if s.strip()]



"""Add featured_asset_id to pages.

Revision ID: m7n8o9p0q1r2
Revises: a1b2c3d4e5f6
Create Date: 2026-02-27
"""

from alembic import op
import sqlalchemy as sa

revision = "m7n8o9p0q1r2"
down_revision = "a1b2c3d4e5f6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("pages") as batch_op:
        batch_op.add_column(
            sa.Column("featured_asset_id", sa.Uuid(), nullable=True),
        )
        batch_op.create_foreign_key(
            "fk_pages_featured_asset_id",
            "assets",
            ["featured_asset_id"],
            ["id"],
            ondelete="SET NULL",
        )


def downgrade() -> None:
    with op.batch_alter_table("pages") as batch_op:
        batch_op.drop_constraint("fk_pages_featured_asset_id", type_="foreignkey")
        batch_op.drop_column("featured_asset_id")
